@@ -1,7 +1,7 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from decimal import Decimal
-from collections import namedtuple
+from collections import namedtuple, defaultdict
 from itertools import groupby
 
 from sql import Null
@@ -13,14 +13,15 @@ from trytond.model import Workflow, ModelView, ModelSQL, fields, Check, \
     sequence_ordered, DictSchemaMixin
 from trytond.pyson import Eval, If, Bool
 from trytond.transaction import Transaction
-from trytond import backend
 from trytond.pool import Pool
 from trytond.wizard import Wizard, StateView, StateAction, Button
 from trytond.modules.company import CompanyReport
 
 __all__ = ['Statement', 'Line', 'LineGroup',
     'Origin', 'OriginInformation',
-    'ImportStatementStart', 'ImportStatement', 'StatementReport']
+    'ImportStatementStart', 'ImportStatement',
+    'ReconcileStatement',
+    'StatementReport']
 
 _STATES = {'readonly': Eval('state') != 'draft'}
 _DEPENDS = ['state']
@@ -131,6 +132,8 @@ class Statement(Workflow, ModelSQL, ModelView):
     state = fields.Selection(STATES, 'State', readonly=True, select=True)
     validation = fields.Function(fields.Char('Validation'),
         'on_change_with_validation')
+    to_reconcile = fields.Function(
+        fields.Boolean("To Reconcile"), 'get_to_reconcile')
 
     @classmethod
     def __setup__(cls):
@@ -144,8 +147,6 @@ class Statement(Workflow, ModelSQL, ModelView):
                     'deletion.'),
                 'paid_invoice_draft_statement': ('There are paid invoices on '
                     'draft statements.'),
-                'debit_credit_account_statement_journal': ('Please provide '
-                    'debit and credit account on statement journal "%s".'),
                 'post_with_pending_amount': ('Origin line "%(origin)s" '
                     'of statement "%(statement)s" still has a pending amount '
                     'of "%(amount)s".'),
@@ -160,45 +161,35 @@ class Statement(Workflow, ModelSQL, ModelView):
         cls._buttons.update({
                 'draft': {
                     'invisible': Eval('state') != 'cancel',
+                    'depends': ['state'],
                     },
                 'validate_statement': {
                     'invisible': Eval('state') != 'draft',
+                    'depends': ['state'],
                     },
                 'post': {
                     'invisible': Eval('state') != 'validated',
+                    'depends': ['state'],
                     },
                 'cancel': {
                     'invisible': ~Eval('state').in_(['draft', 'validated']),
+                    'depends': ['state'],
+                    },
+                'reconcile': {
+                    'invisible': Eval('state').in_(['draft', 'cancel']),
+                    'readonly': ~Eval('to_reconcile'),
+                    'depends': ['state', 'to_reconcile'],
                     },
                 })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         transaction = Transaction()
         cursor = transaction.connection.cursor()
         sql_table = cls.__table__()
 
-        # Migration from 1.8: new field company
-        table = TableHandler(cls, module_name)
-        company_exist = table.column_exist('company')
-
         super(Statement, cls).__register__(module_name)
-
-        # Migration from 1.8: fill new field company
-        if not company_exist:
-            offset = 0
-            limit = transaction.database.IN_MAX
-            statements = True
-            while statements:
-                statements = cls.search([], offset=offset, limit=limit)
-                offset += limit
-                for statement in statements:
-                    cls.write([statement], {
-                            'company': statement.journal.company.id,
-                            })
-            table = TableHandler(cls, module_name)
-            table.not_null_action('company', action='add')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 3.2: remove required on start/end balance
         table.not_null_action('start_balance', action='remove')
@@ -264,63 +255,80 @@ class Statement(Workflow, ModelSQL, ModelView):
         return ((getattr(self, 'end_balance', 0) or 0)
             - (getattr(self, 'start_balance', 0) or 0))
 
-    @fields.depends('lines', 'journal')
+    @fields.depends('lines', 'journal', 'company')
     def on_change_lines(self):
         pool = Pool()
-        Currency = pool.get('currency.currency')
         Line = pool.get('account.statement.line')
-        if self.journal and self.lines:
-            invoices = set()
-            for line in self.lines:
-                if getattr(line, 'invoice', None):
-                    invoices.add(line.invoice)
-            invoice_id2amount_to_pay = {}
-            for invoice in invoices:
-                with Transaction().set_context(date=invoice.currency_date):
-                    if invoice.type == 'out':
-                        sign = -1
-                    else:
-                        sign = 1
-                    invoice_id2amount_to_pay[invoice.id] = sign * (
-                        Currency.compute(invoice.currency,
-                            invoice.amount_to_pay, self.journal.currency))
+        if not self.journal or not self.lines or not self.company:
+            return
+        if self.journal.currency != self.company.currency:
+            return
 
-            lines = list(self.lines)
-            line_offset = 0
-            for index, line in enumerate(self.lines or []):
-                if getattr(line, 'invoice', None) and line.id:
-                    amount_to_pay = invoice_id2amount_to_pay[line.invoice.id]
-                    if (not self.journal.currency.is_zero(amount_to_pay)
-                            and getattr(line, 'amount', None)
-                            and (line.amount >= 0) == (amount_to_pay <= 0)):
-                        if abs(line.amount) > abs(amount_to_pay):
-                            new_line = Line()
-                            for field_name, field in Line._fields.iteritems():
-                                if field_name == 'id':
-                                    continue
-                                try:
-                                    setattr(new_line, field_name,
-                                        getattr(line, field_name))
-                                except AttributeError:
-                                    pass
-                            new_line.amount = line.amount + amount_to_pay
-                            new_line.invoice = None
-                            line_offset += 1
-                            lines.insert(index + line_offset, new_line)
-                            invoice_id2amount_to_pay[line.invoice.id] = 0
-                            line.amount = amount_to_pay.copy_sign(line.amount)
-                        else:
-                            invoice_id2amount_to_pay[line.invoice.id] = (
-                                line.amount + amount_to_pay)
+        invoices = set()
+        for line in self.lines:
+            if (getattr(line, 'invoice', None)
+                    and line.invoice.currency == self.company.currency):
+                invoices.add(line.invoice)
+        invoice_id2amount_to_pay = {}
+        for invoice in invoices:
+            if invoice.type == 'out':
+                sign = -1
+            else:
+                sign = 1
+            invoice_id2amount_to_pay[invoice.id] = sign * invoice.amount_to_pay
+
+        lines = list(self.lines)
+        line_offset = 0
+        for index, line in enumerate(self.lines or []):
+            if getattr(line, 'invoice', None) and line.id:
+                if line.invoice.id not in invoice_id2amount_to_pay:
+                    continue
+                amount_to_pay = invoice_id2amount_to_pay[line.invoice.id]
+                if (amount_to_pay
+                        and getattr(line, 'amount', None)
+                        and (line.amount >= 0) == (amount_to_pay <= 0)):
+                    if abs(line.amount) > abs(amount_to_pay):
+                        new_line = Line()
+                        for field_name, field in Line._fields.items():
+                            if field_name == 'id':
+                                continue
+                            try:
+                                setattr(new_line, field_name,
+                                    getattr(line, field_name))
+                            except AttributeError:
+                                pass
+                        new_line.amount = line.amount + amount_to_pay
+                        new_line.invoice = None
+                        line_offset += 1
+                        lines.insert(index + line_offset, new_line)
+                        invoice_id2amount_to_pay[line.invoice.id] = 0
+                        line.amount = amount_to_pay.copy_sign(line.amount)
                     else:
-                        line.invoice = None
-            self.lines = lines
+                        invoice_id2amount_to_pay[line.invoice.id] = (
+                            line.amount + amount_to_pay)
+                else:
+                    line.invoice = None
+        self.lines = lines
 
     @fields.depends('journal')
     def on_change_with_validation(self, name=None):
         if self.journal:
             return self.journal.validation
 
+    def get_to_reconcile(self, name=None):
+        return bool(self.lines_to_reconcile)
+
+    @property
+    def lines_to_reconcile(self):
+        lines = []
+        for line in self.lines:
+            if line.move:
+                for move_line in line.move.lines:
+                    if (move_line.account.reconcile
+                            and not move_line.reconciliation):
+                        lines.append(move_line)
+        return lines
+
     def _group_key(self, line):
         key = (
             ('number', line.number or Unequal()),
@@ -382,12 +390,8 @@ class Statement(Workflow, ModelSQL, ModelView):
         end_balance = (self.start_balance
             + sum(l.amount for l in self.lines))
         if end_balance != self.end_balance:
-            lang, = Lang.search([
-                    ('code', '=', Transaction().language),
-                    ])
-            amount = Lang.format(lang,
-                '%.' + str(self.journal.currency.digits) + 'f',
-                end_balance, True)
+            lang = Lang.get()
+            amount = lang.currency(end_balance, self.journal.currency)
             self.raise_user_error('wrong_end_balance', amount)
 
     def validate_amount(self):
@@ -396,12 +400,8 @@ class Statement(Workflow, ModelSQL, ModelView):
 
         amount = sum(l.amount for l in self.lines)
         if amount != self.total_amount:
-            lang, = Lang.search([
-                    ('code', '=', Transaction().language),
-                    ])
-            amount = Lang.format(lang,
-                '%.' + str(self.journal.currency.digits) + 'f',
-                amount, True)
+            lang = Lang.get()
+            amount = lang.currency(amount, self.journal.currency)
             self.raise_user_error('wrong_total_amount', amount)
 
     def validate_number_of_lines(self):
@@ -484,9 +484,7 @@ class Statement(Workflow, ModelSQL, ModelView):
 
         MoveLine.save([l for l, _ in move_lines])
 
-        for move_line, line in move_lines:
-            if line:
-                line.reconcile(move_line)
+        Line.reconcile(move_lines)
         return moves
 
     def _get_move(self, key):
@@ -502,7 +500,7 @@ class Statement(Workflow, ModelSQL, ModelView):
             date=key['date'],
             origin=self,
             company=self.company,
-            description=unicode(key['number']),
+            description=str(key['number']),
             )
 
     def _get_move_line(self, amount, amount_second_currency, lines):
@@ -510,15 +508,6 @@ class Statement(Workflow, ModelSQL, ModelView):
         pool = Pool()
         MoveLine = pool.get('account.move.line')
 
-        if amount < 0:
-            account = self.journal.journal.debit_account
-        else:
-            account = self.journal.journal.credit_account
-
-        if not account:
-            self.raise_user_error('debit_credit_account_statement_journal',
-                (self.journal.rec_name,))
-
         if self.journal.currency != self.company.currency:
             second_currency = self.journal.currency
             amount_second_currency *= -1
@@ -535,7 +524,7 @@ class Statement(Workflow, ModelSQL, ModelView):
         return MoveLine(
             debit=abs(amount) if amount < 0 else 0,
             credit=abs(amount) if amount > 0 else 0,
-            account=account,
+            account=self.journal.account,
             second_currency=second_currency,
             amount_second_currency=amount_second_currency,
             description=description,
@@ -567,6 +556,11 @@ class Statement(Workflow, ModelSQL, ModelView):
         lines = [l for s in statements for l in s.lines]
         StatementLine.delete_move(lines)
 
+    @classmethod
+    @ModelView.button_action('account_statement.act_reconcile')
+    def reconcile(cls, statements):
+        pass
+
 
 def origin_mixin(_states, _depends):
     class Mixin:
@@ -596,7 +590,10 @@ def origin_mixin(_states, _depends):
                 ('company', '=', Eval('company', 0)),
                 ('kind', '!=', 'view'),
                 ],
-            states=_states, depends=_depends + ['company'])
+            context={
+                'date': Eval('date'),
+                },
+            states=_states, depends=_depends + ['company', 'date'])
         description = fields.Char(
             "Description", states=_states, depends=_depends)
 
@@ -671,10 +668,6 @@ class Line(
                     })
             cls.date.depends.append('origin')
         cls.account.required = True
-        cls._error_messages.update({
-                'amount_greater_invoice_amount_to_pay': ('Amount "%s" is '
-                    'greater than the amount to pay of invoice.'),
-                })
         t = cls.__table__()
         cls._sql_constraints += [
             ('check_statement_line_amount', Check(t, t.amount != 0),
@@ -685,14 +678,15 @@ class Line(
     def default_amount():
         return Decimal(0)
 
-    @fields.depends('amount', 'party', 'invoice')
+    @fields.depends('amount', 'party', 'invoice', 'date')
     def on_change_party(self):
         if self.party:
             if self.amount:
-                if self.amount > Decimal("0.0"):
-                    self.account = self.party.account_receivable_used
-                else:
-                    self.account = self.party.account_payable_used
+                with Transaction().set_context(date=self.date):
+                    if self.amount > Decimal("0.0"):
+                        self.account = self.party.account_receivable_used
+                    else:
+                        self.account = self.party.account_payable_used
 
         if self.invoice:
             if self.party:
@@ -702,20 +696,21 @@ class Line(
                 self.invoice = None
 
     @fields.depends('amount', 'party', 'account', 'invoice', 'statement',
-        '_parent_statement.journal')
+        'date', '_parent_statement.journal')
     def on_change_amount(self):
         Currency = Pool().get('currency.currency')
         if self.party:
-            if self.account and self.account not in (
-                    self.party.account_receivable_used,
-                    self.party.account_payable_used):
-                # The user has entered a non-default value, we keep it.
-                pass
-            elif self.amount:
-                if self.amount > Decimal("0.0"):
-                    self.account = self.party.account_receivable_used
-                else:
-                    self.account = self.party.account_payable_used
+            with Transaction().set_context(date=self.date):
+                if self.account and self.account not in (
+                        self.party.account_receivable_used,
+                        self.party.account_payable_used):
+                    # The user has entered a non-default value, we keep it.
+                    pass
+                elif self.amount:
+                    if self.amount > Decimal("0.0"):
+                        self.account = self.party.account_receivable_used
+                    else:
+                        self.account = self.party.account_payable_used
         if self.invoice:
             if self.amount and self.statement and self.statement.journal:
                 invoice = self.invoice
@@ -750,7 +745,7 @@ class Line(
         '_parent_origin.party', '_parent_origin.account',
         '_parent_origin.number', '_parent_origin.description',
         '_parent_origin.statement',
-        methods=['party'])
+        methods=['on_change_party'])
     def on_change_origin(self):
         if self.origin:
             self.amount = self.origin.pending_amount
@@ -795,49 +790,43 @@ class Line(
     def copy(cls, lines, default=None):
         if default is None:
             default = {}
-        default = default.copy()
+        else:
+            default = default.copy()
         default.setdefault('move', None)
         default.setdefault('invoice', None)
         return super(Line, cls).copy(lines, default=default)
 
-    def reconcile(self, move_line):
+    @classmethod
+    def reconcile(cls, move_lines):
         pool = Pool()
-        Currency = pool.get('currency.currency')
-        Lang = pool.get('ir.lang')
         Invoice = pool.get('account.invoice')
         MoveLine = pool.get('account.move.line')
 
-        if self.invoice:
-            with Transaction().set_context(date=self.invoice.currency_date):
-                amount_to_pay = Currency.compute(self.invoice.currency,
-                    self.invoice.amount_to_pay,
-                    self.statement.journal.currency)
-            if abs(amount_to_pay) < abs(self.amount):
-                lang, = Lang.search([
-                        ('code', '=', Transaction().language),
-                        ])
+        invoice_payments = defaultdict(list)
 
-                amount = Lang.format(lang,
-                    '%.' + str(self.statement.journal.currency.digits) + 'f',
-                    self.amount, True)
-                self.raise_user_error('amount_greater_invoice_amount_to_pay',
-                        error_args=(amount,))
+        to_reconcile = []
+        for move_line, line in move_lines:
+            if not line or not line.invoice:
+                continue
 
-            with Transaction().set_context(date=self.invoice.currency_date):
-                amount = Currency.compute(self.statement.journal.currency,
-                    self.amount, self.statement.company.currency)
+            # Write previous invoice payments to have them when calling
+            # get_reconcile_lines_for_amount
+            if line.invoice in invoice_payments:
+                Invoice.add_payment_lines(invoice_payments)
+                invoice_payments.clear()
 
-            reconcile_lines = self.invoice.get_reconcile_lines_for_amount(
-                amount)
+            reconcile_lines = line.invoice.get_reconcile_lines_for_amount(
+                move_line.credit - move_line.debit)
 
-            assert move_line.account == self.invoice.account
+            assert move_line.account == line.invoice.account
 
-            Invoice.write([self.invoice], {
-                    'payment_lines': [('add', [move_line.id])],
-                    })
-            if reconcile_lines[1] == Decimal('0.0'):
-                lines = reconcile_lines[0] + [move_line]
-                MoveLine.reconcile(lines)
+            invoice_payments[line.invoice].append(move_line.id)
+            if not reconcile_lines[1]:
+                to_reconcile.append(reconcile_lines[0] + [move_line])
+        if invoice_payments:
+            Invoice.add_payment_lines(invoice_payments)
+        if to_reconcile:
+            MoveLine.reconcile(*to_reconcile)
 
     @classmethod
     def post_move(cls, lines):
@@ -851,8 +840,8 @@ class Line(
         Reconciliation = pool.get('account.move.reconciliation')
 
         reconciliations = [l.reconciliation
-            for line in lines for l in line.move.lines
-            if line.move and l.reconciliation]
+            for line in lines if line.move
+            for l in line.move.lines if l.reconciliation]
         Reconciliation.delete(reconciliations)
         Move.delete(list({l.move for l in lines if l.move}))
 
@@ -1062,11 +1051,28 @@ class ImportStatement(Wizard):
             statement.origin_file = fields.Binary.cast(self.start.file_)
         Statement.save(statements)
 
-        data = {'res_id': map(int, statements)}
+        data = {'res_id': list(map(int, statements))}
         if len(statements) == 1:
             action['views'].reverse()
         return action, data
 
 
+class ReconcileStatement(Wizard):
+    "Statement Reconcile"
+    __name__ = 'account.statement.reconcile'
+    start = StateAction('account.act_reconcile')
+
+    def do_start(self, action):
+        pool = Pool()
+        Statement = pool.get('account.statement')
+        statements = Statement.browse(Transaction().context['active_ids'])
+        lines = sum(([int(l) for l in s.lines_to_reconcile]
+                for s in statements), [])
+        return action, {
+            'model': 'account.move.line',
+            'ids': lines,
+            }
+
+
 class StatementReport(CompanyReport):
     __name__ = 'account.statement'
